-module(wsEgHer).

-behaviour(wsHer).

-include("wsCom.hrl").
-include_lib("kernel/include/file.hrl").

-export([
   handle/3
]).

-export([
   chunk_loop/1
]).

-spec handle(Method :: method(), Path :: path(), WsReq :: wsReq()) -> wsHer:response().
handle('GET', <<"/hello/world">>, WsReq) ->
   io:format("IMY************XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX ~n receive WsReq: ~p~n", [WsReq]),
   %% Reply with a normal response.
   %timer:sleep(1000),
   {ok, [], <<"Hello World!">>};

handle('GET', <<"/hello">>, WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Fetch a GET argument from the URL.
   Name = wsReq:get_arg(<<"name">>, WsReq, <<"undefined">>),
   {ok, [], <<"Hello ", Name/binary>>};

handle('POST', <<"hello">>, WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Fetch a POST argument from the POST body.
   Name = wsReq:post_arg(<<"name">>, WsReq, <<"undefined">>),
   %% Fetch and decode
   City = wsReq:post_arg_decoded(<<"city">>, WsReq, <<"undefined">>),
   {ok, [], <<"Hello ", Name/binary, " of ", City/binary>>};

handle('GET', [<<"hello">>, <<"iolist">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Iolists will be kept as iolists all the way to the socket.
   Name = wsReq:get_arg(<<"name">>, WsReq),
   {ok, [], [<<"Hello ">>, Name]};

handle('GET', [<<"type">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   Name = wsReq:get_arg(<<"name">>, WsReq),
   %% Fetch a header.
   case wsReq:get_header(<<"Accept">>, WsReq, <<"text/plain">>) of
      <<"text/plain">> ->
         {ok, [{<<"content-type">>, <<"text/plain; charset=ISO-8859-1">>}],
            <<"name: ", Name/binary>>};
      <<"application/json">> ->
         {ok, [{<<"content-type">>,
            <<"application/json; charset=ISO-8859-1">>}],
            <<"{\"name\" : \"", Name/binary, "\"}">>}
   end;

handle('GET', [<<"headers.html">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Set custom headers, for example 'Content-Type'
   {ok, [{<<"X-Custom">>, <<"foobar">>}], <<"see headers">>};

%% See note in function doc re: overriding Elli's default behaviour
%% via Connection and Content-Length headers.
handle('GET', [<<"user">>, <<"defined">>, <<"behaviour">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {304, [{<<"Connection">>, <<"close">>},
      {<<"Content-Length">>, <<"123">>}], <<"ignored">>};

handle('GET', [<<"user">>, <<"content-length">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {200, [{<<"Content-Length">>, 123}], <<"foobar">>};

handle('GET', [<<"crash">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Throwing an exception results in a 500 response and
   %% request_throw being called
   throw(foobar);

handle('GET', [<<"decoded-hello">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Fetch a URI decoded GET argument from the URL.
   Name = wsReq:get_arg_decoded(<<"name">>, WsReq, <<"undefined">>),
   {ok, [], <<"Hello ", Name/binary>>};

handle('GET', [<<"decoded-list">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Fetch a URI decoded GET argument from the URL.
   [{<<"name">>, Name}, {<<"foo">>, true}] =
      wsReq:get_args_decoded(WsReq),
   {ok, [], <<"Hello ", Name/binary>>};


handle('GET', [<<"sendfile">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Returning {file, "/path/to/file"} instead of the body results
   %% in Elli using sendfile.
   F = "README.md",
   {ok, [], {file, F}};

handle('GET', [<<"send_no_file">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Returning {file, "/path/to/file"} instead of the body results
   %% in Elli using sendfile.
   F = "README",
   {ok, [], {file, F}};

handle('GET', [<<"sendfile">>, <<"error">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   F = "test",
   {ok, [], {file, F}};

handle('GET', [<<"sendfile">>, <<"range">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Read the Range header of the request and use the normalized
   %% range with sendfile, otherwise send the entire file when
   %% no range is present, or respond with a 416 if the range is invalid.
   F = "README.md",
   {ok, [], {file, F, wsReq:get_range(WsReq)}};

handle('GET', [<<"compressed">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Body with a byte size over 1024 are automatically gzipped by
   %% elli_middleware_compress
   {ok, binary:copy(<<"Hello World!">>, 86)};

handle('GET', [<<"compressed-io_list">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Body with a iolist size over 1024 are automatically gzipped by
   %% elli_middleware_compress
   {ok, lists:duplicate(86, [<<"Hello World!">>])};


handle('HEAD', [<<"head">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {200, [], <<"body must be ignored">>};

handle('GET', [<<"chunked">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Start a chunked response for streaming real-time events to the
   %% browser.
   %%
   %% Calling elli_request:send_chunk(ChunkRef, Body) will send that
   %% part to the client. elli_request:close_chunk(ChunkRef) will
   %% close the response.
   %%
   %% Return immediately {chunk, Headers} to signal we want to chunk.
   Ref = wsReq:chunk_ref(WsReq),
   spawn(fun() -> ?MODULE:chunk_loop(Ref) end),
   {chunk, [{<<"Content-Type">>, <<"text/event-stream">>}]};

handle('GET', [<<"shorthand">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {200, <<"hello">>};

handle('GET', [<<"ip">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {<<"200 OK">>, wsReq:peer(WsReq)};

handle('GET', [<<"304">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% A "Not Modified" response is exactly like a normal response (so
   %% Content-Length is included), but the body will not be sent.
   {304, [{<<"Etag">>, <<"foobar">>}], <<"Ignored">>};

handle('GET', [<<"302">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {302, [{<<"Location">>, <<"/hello/world">>}], <<>>};

handle('GET', [<<"403">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   %% Exceptions formatted as return codes can be used to
   %% short-circuit a response, for example in case of
   %% authentication/authorization
   throw({403, [], <<"Forbidden">>});

handle('GET', [<<"invalid_return">>], WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {invalid_return};

handle(_, _, WsReq) ->
   io:format("receive WsReq: ~p~n", [WsReq]),
   {404, [], <<"Not Found">>}.


%% @doc Send 10 separate chunks to the client.
%% @equiv chunk_loop(Ref, 10)
chunk_loop(Ref) ->
   chunk_loop(Ref, 10).

%% @doc If `N > 0', send a chunk to the client, checking for errors,
%% as the user might have disconnected.
%% When `N == 0', call {@link elli_request:close_chunk/1.
%% elli_request:close_chunk(Ref)}.
chunk_loop(Ref, 0) ->
   wsReq:close_chunk(Ref);
chunk_loop(Ref, N) ->
   timer:sleep(10),

   case wsReq:send_chunk(Ref, [<<"chunk">>, integer_to_binary(N)]) of
      ok -> ok;
      {error, Reason} -> ?wsErr("error in sending chunk: ~p~n", [Reason])
   end,
   chunk_loop(Ref, N - 1).